#include "utils.h"
#include "massb.h"
#include "sa_search.h"
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_sort_vector.h>
#include <mpi.h>

int main ( int argc, char *argv[]  ) {

	MPI::Init();
	int nproc = MPI::COMM_WORLD.Get_size();
	int tid = MPI::COMM_WORLD.Get_rank();

	int master = 0;	// master thread is 0
	
	// Each thread (cpu) opens and makes it's own copies of MASSB P and SABS.
	MASSB P;
	SA_search SABS( P );
	SABS.cpuid = tid;
	P.cpuid = tid;
	
	if (tid == master) {
		char hname[12];
  		gethostname(hname,12);
		printf("master is tid %i on %s\n",tid,hname);

		int check = 255;
		for (int i=1; i<argc; i+=2) {
			if ( !strcmp(argv[i],"MASSBFILE") ) check -= 128;
			else if ( !strcmp(argv[i],"ETAT" ) ) check -= 64;
			else if ( !strcmp(argv[i],"MINTEMP" ) ) check -= 32;
			else if ( !strcmp(argv[i],"INITTEMP" ) ) check -= 16;
			else if ( !strcmp(argv[i],"MAXSTEP" ) ) check -= 8;
			else if ( !strcmp(argv[i],"BOLTZMANN" ) ) check -= 4;
			else if ( !strcmp(argv[i],"ATTENTEMP" ) ) check -= 2;
			else if ( !strcmp(argv[i],"ITERT" ) ) check -= 1;
		}
		if (check != 0) {
			cerr << "[bs_sa_mpi] Need to specify:" << endl;
			cerr << "MASSBFILE" << endl;
			cerr << "ETAT (integer for new state)" << endl;
			cerr << "INITTEMP (initial temperature)" << endl;
			cerr << "MINTEMP (temperature at which to stop)" << endl;
			cerr << "ATTENTEMP (factor to reduce temp at each step)" << endl;
			cerr << "MAXSTEP (maximum random step to take)" << endl;
			cerr << "BOLTZMANN (effective boltzmann constant)" << endl;
			cerr << "ITERT (number of steps to take per temperature)" << endl;
			exit(0);
		}
		P.append_logfile(); 
		SABS.lfile = &P.lfile;
	}

	SABS.parseinputs( argc, argv );

	strcpy( P.bsetatfile, SABS.massbfile );
	P.etat = SABS.etat;
//	for (int p=0; p<nproc; p++) {
		P.read_full_bsetat();
//	}

	SABS.Lagrange_siman( P );

	P.write_ndata();
	P.write_dist();
	// gather chi2, entropy, and lagrange multipliers at master
	double *lagrangeconfig = new double [2+P.datanum];	
	double *gathered = new double [ nproc * (2+P.datanum) ];

	lagrangeconfig[0] = P.chi2;
	lagrangeconfig[1] = P.entropy;
	for (int ell=0; ell<P.datanum; ell++) {
		lagrangeconfig[ell+2] = gvg(P.lagrange,ell);
	}

	MPI::Datatype lagrangeconfig_mpi;
	lagrangeconfig_mpi = MPI::DOUBLE.Create_contiguous( 2+P.datanum );
	lagrangeconfig_mpi.Commit();
	MPI::COMM_WORLD.Gather(lagrangeconfig,1,lagrangeconfig_mpi,
		gathered,1,lagrangeconfig_mpi,master);

	if (tid == master) {
		char laconfname[81];
		sprintf(laconfname,"%s/%s/%s.la.%i",
			P.outpath,P.outbase,P.outbase,P.etat);
		ofstream laconf;
		laconf.open( laconfname, ios::out );
		if ( !laconf.is_open() )	{
			cout << "Can't open laconfig file " << laconfname << endl;
			exit(0);
		}
		
	// rank final states from each processor, on chi2
		char num[20];
		gsl_vector *vs = gsl_vector_calloc(nproc);
		gsl_permutation *p = gsl_permutation_calloc(nproc);
		for (int c=0; c<nproc; c++) {
			gvs( vs,c, gathered[c*(2+P.datanum)] );
		}
		gsl_sort_vector_index(p,vs);

		for (int c=0; c<nproc; c++) {
			int ii = gsl_permutation_get(p,c);
			for (int la=0; la<2+P.datanum; la++) {
				sprintf(num,"%20g ",gathered[ ii*(2+P.datanum)+la ]);
				laconf << num;
			}
			laconf << endl;
		}
		laconf.close();
	}
			
	MPI_Finalize();
	return(0);
}
