#include "massb.h"
#include <mpi.h>

int main ( int argc, char *argv[]  ) {

	MPI::Init();
	int nproc = MPI::COMM_WORLD.Get_size();
	int tid = MPI::COMM_WORLD.Get_rank();
	
	int master = 0;	// master thread is 0
	int * dims;
	double * mparam;
	double * gathered;

	if (tid == master) {
		char hname[12];
  		gethostname(hname,12);
		printf("master is tid %i on %s\n",tid,hname);

		int check = 7;
		for (int i=1; i<argc; i+=2) {
			if ( !strcmp(argv[i],"PP2PI") ) check -= 4;
			else if ( !strcmp(argv[i],"DATAFILE" ) ) check -= 2;
			else if ( !strcmp(argv[i],"DISTFILE" ) ) check -= 1;
		}
		if (check != 0) {
			cerr << "[calc_matrix_mpi] Need to specify:" << endl;
			cerr << "DATAFILE" << endl;
			cerr << "DISTFILE" << endl;
			cerr << "PP2PI" << endl;
			exit(0);
		}

		MASSB P;
		P.parseinputs( argc, argv );
		
		// read ndata and dist files and 
		P.read_ndata();
		P.read_dist();
		P.get_datadist_md5();
		
		// mparam is [dist-param-1][dist-param-2][dist-param-...][data-param-1]
		//		[data-param-2][data-param-...]
		mparam = new double [P.distnum*P.distdim + P.datanum*P.datadim];	
		dims = new int [5];

		for (int p=0; p<P.distdim; p++) {
			for (int n=0; n<P.distnum; n++) {
				mparam[p*P.distnum+n] = gmg(P.dist_params,n,p);
			}
		}
		for (int p=0; p<P.datadim; p++) {
			for (int n=0; n<P.datanum; n++) {
				mparam[P.distnum*P.distdim + p*P.datanum+n] = gmg(P.data_params,n,p);
			}
		}
		dims[0] = P.pp2pi;
		dims[1] = P.distnum;
		dims[2] = P.distdim;
		dims[3] = P.datanum;
		dims[4] = P.datadim;

		// allocate memory for vector to be gathered to master from slaves
		gathered = new double [nproc * P.distnum * P.datanum];

		//P.T.mark(&P.lfile,"MPI: about to begin beta looping ");
		
		time_t seconds = time(NULL);
		P.lfile << "MPI: begin beta loop "  << ctime(&seconds) << endl;
		P.lfile << "nested loop combinations: " << pow(P.pp2pi,3)/2 << endl << flush;				
		P.lfile << "!DATADIST_MD5CHECK " << P.datadist_md5 << endl;
	} else {
		dims = new int [5];
	}

	// send pp2pi, distnum, distdim, datanum, datadim to all processors
	// so they can prepare a data structure to receive the distribution
	// and data configurations 
	MPI::Datatype ldims;
	ldims = MPI::INT.Create_contiguous( 5 );
	ldims.Commit();
	MPI::COMM_WORLD.Bcast( dims,1,ldims,master);

	if (tid != master) {
		mparam = new double [dims[1]*dims[2] + dims[3]*dims[4] ];
	}

	// create and broadcast the data and distribution parameter
	// configurations to all processors
	MPI::Datatype linearparams;
	linearparams = 
		MPI::DOUBLE.Create_contiguous( dims[1]*dims[2] + dims[3]*dims[4] );
	linearparams.Commit();
	MPI::COMM_WORLD.Bcast( mparam,1,linearparams,master);

	// prepare MPI datastructure for returning each processors work
	// back to the master
	MPI::Datatype localwork;
	localwork = MPI::DOUBLE.Create_contiguous( dims[1]*dims[3] );
	localwork.Commit();

	// Work out what values of beta should be run on each process
	// by calculating min and max values. All processors receive the
	// same amount of work -- if there aren't a multiple-of-nproc 
	// number of beta values, 1 additional beta value is assigned to
	// the first few processes.
	int min, max;
	int mxi = int( double(dims[0]/2)/double(nproc));
	int mxi1 = dims[0]/2 - mxi*nproc;

	if (tid < mxi1) {
		min = tid * (1 + mxi);
		max = min + mxi;
	} else {
		min = tid * mxi + mxi1;
		max = min + mxi - 1;
	}

	// Report process tid, hostname, min, max ... send in message back to
	// master for logging to lfile
	char message[100];
	char hname[12];
  	gethostname(hname,12);
	
	int tag = 113;
	MPI_Status	status; // return status for receive
	sprintf(message,
		"Process = %i of %i on %s: min %i max %i\n",
		tid, nproc, hname, min, max);
//	printf("%s\n",message);
	MPI_Send(message,strlen(message)+1,MPI_CHAR,0,tag,MPI_COMM_WORLD);
	if (tid == 0) {
		MASSB P;
		for (int source = 0; source < nproc; source++) {
			MPI_Recv(message,100,MPI_CHAR,source,tag,MPI_COMM_WORLD,&status);
//			cout << message << flush;
			P.lfile << message << flush;
		}
	}

	// each process fills its own linear version of a
	// dist2data matrix, which will be summed together
	// at the master when all processes are finished.
	double * l_dist2data = new double [dims[1] * dims[3] ];
	for (int v=0; v<dims[1]*dims[3]; v++) {
		l_dist2data[v] = 0.0;
	}

	// Each process calculates its own trig vectors
	gsl_vector * sinangle = gsl_vector_alloc(dims[0]);	
	gsl_vector * cosangle = gsl_vector_alloc(dims[0]);	
	gsl_vector * sindangle = gsl_vector_alloc(dims[0]);	
	gsl_vector * cosdangle = gsl_vector_alloc(dims[0]);	
	
 	double incpi = 6.28318530718/double(dims[0]);
	for (int j=0; j<dims[0]; j++) {
		double angle = double(j)*incpi;
		gvs(sinangle,j,sin(angle) );
		gvs(cosangle,j,cos(angle) );
		gvs(sindangle,j,sin(2.0*angle) );
		gvs(cosdangle,j,cos(2.0*angle) );
	}

	double sinalpha,sinbeta,singamma,cosalpha,cosbeta,cosgamma;
	double sin2alpha,sin2beta,sin2gamma,cos2alpha,cos2beta,cos2gamma;
	double PhiA,PhiB,Phitmp,Phitmp2,Phi_Nprime,Iab,csum,ssum;

	double delta, eta, mas, N;
	gsl_matrix * Csum = gsl_matrix_calloc(dims[3],dims[1]); 
	gsl_matrix * Ssum = gsl_matrix_calloc(dims[3],dims[1]); 

	double * distpar = new double [ dims[1] ];
	double * datapar = new double [ dims[3] ];
	int totdistpar = dims[1]*dims[2];

	// integrate over beta, each process knows which bracket
	// (min-max) to work on.
	for (int b=min; b<=max; b++) {
//		T.simpletimestamp(&lfile);
//		T.reportthis_coreduty( &lfile );
//			T.incr_coreduty();
		sinbeta = gvg(sinangle,b);
		sin2beta = gvg(sindangle,b);
		cosbeta = gvg(cosangle,b);
		cos2beta = gvg(cosdangle,b);

		// integrate over alpha
		for (int a=0; a<dims[0]; a++) {
			sinalpha = gvg(sinangle,a);
			sin2alpha = gvg(sindangle,a);
			cosalpha = gvg(cosangle,a);
			cos2alpha = gvg(cosdangle,a);
			
			gsl_matrix_set_all(Csum,0.0);	
			gsl_matrix_set_all(Ssum,0.0);	

			for (int g=0; g<dims[0]; g++) {
				double gamma = double(g)*incpi;

				singamma = gvg(sinangle,g);
				sin2gamma = gvg(sindangle,g);
				cosgamma = gvg(cosangle,g);
				cos2gamma = gvg(cosdangle,g);

				PhiA = cos2alpha*(3.0+cos2beta)*sin2gamma/(-24.0);
				PhiA -= 0.235702260395516 * cos2alpha*sin2beta*singamma;
				PhiA -= 0.471404520791032 * sin2alpha*sinbeta*cosgamma;
				PhiA -= 0.166666666666667 * sin2alpha*cosbeta*cos2gamma;

				PhiB = (1.0 - cos2beta)*sin2gamma/8.0;
				PhiB -= 0.707106781186547*sin2beta*singamma;

				for (int cm=0; cm<dims[1]; cm++) {
					// make this complete general first ...
					for (int distd=0; distd<dims[2]; distd++) {
						distpar[distd] = mparam[distd*dims[1] + cm];
					}
					// then assign names meaningful for application
					delta = distpar[0];
					eta   = distpar[1];

					Phitmp = delta*eta*PhiA + delta*PhiB;
	
					for (int n=0; n<dims[3]; n++) {
						// make this complete general first ...
						for (int datad=0; datad<dims[4]; datad++) {
							datapar[datad] = 
								mparam[totdistpar + datad*dims[3] + n];
						}
						// then assign names meaningful for application
						mas = datapar[0];
						N   = datapar[1];

						Phitmp2 = Phitmp/mas;
						Phi_Nprime = N*gamma - Phitmp2;
						csum = gmg(Csum,n,cm);
						ssum = gmg(Ssum,n,cm);
						gsl_matrix_set( Csum, n, cm, 
							csum + cos( Phi_Nprime) );
						gsl_matrix_set( Ssum, n, cm, 
							ssum + sin( Phi_Nprime) );
					}
				} // end configurational matrix (delta,eta) block
			} // End gamma block
			for (int cm=0; cm<dims[1]; cm++) {
				for (int n=0; n<dims[3]; n++) {
					Iab = (pow( gmg( Csum, n, cm), 2.0) 
						+ pow( gmg( Ssum, n, cm), 2.0));
					l_dist2data[n*dims[1] + cm] += Iab*sinbeta;
				} // end sideband block
			} // end config matrix (delta,eta) block

		} // End alpha block
	} // End beta block
//		T.report_coreduty( &lfile, pp2pi/2 );

	gsl_matrix_free(Csum);
	gsl_matrix_free(Ssum);

	// all data gathered back at the master
	MPI::COMM_WORLD.Gather(l_dist2data,1,localwork,
		gathered,1,localwork,master);

	delete l_dist2data;

	double s;		
	if ( tid == master ) {
		MASSB P;

		// need to write matrix header stuff in master block above while data
		// md5/ etc is still accessible, then just append matrix stuff here
		P.dist2data = gsl_matrix_calloc( dims[3], dims[1] ); 
		for (int n=0; n<dims[3]; n++) {
			for (int cm=0; cm<dims[1]; cm++) {
				for (int p=0; p<nproc; p++) {
					s = gmg(P.dist2data,n,cm);
					gms(P.dist2data,n,cm, 
						s + gathered[p*dims[3]*dims[1] + n*dims[1] + cm]);
				}
			}
		}
		delete gathered;
		P.distnum = dims[1];
		P.distdim = dims[2];
		P.datanum = dims[3];
		P.datadim = dims[4];
//		P.read_ndata();
//		P.read_dist();

		gsl_matrix_scale(P.dist2data,3.14159265358979/(pow(double(dims[0]),4.0)));

//		P.T.mark(&P.lfile,"finished beta looping of powderopt = 0");
		time_t seconds = time(NULL);
		P.lfile << "MPI: ended beta loop "  << ctime(&seconds) << endl << flush;
		P.write_dist2data_matrix();
		double cumsum = 0;
		for (int cm=0; cm<dims[1]; cm++) 
			for (int n=0; n<dims[3]; n++) 
				cumsum += gmg(P.dist2data,n,cm);
		P.lfile << "cumulative sum: " << cumsum << endl;
		cout << "cumulative sum = " << cumsum << endl;
		}

	gsl_vector_free(sinangle);
	gsl_vector_free(cosangle);
	gsl_vector_free(sindangle);
	gsl_vector_free(cosdangle);

	MPI_Finalize();
	return(0);
}
